#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <immintrin.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/filter.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <math.h>
#include <netdb.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/if.h>
#include <pthread.h>
#include <sched.h>
#include <semaphore.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "private/trx_driver.h"
/* Proprietary code:
     - fast conversion between int16_t and float
*/
#include "private/convert16_sse.c"

#define DEBUG
//#define DISABLE_SEND
//#define DISABLE_WRITE
//#define DISABLE_RECV
//#define DISABLE_READ
//#define DISABLE_STAT
#define DISABLE_RECV2

//#define RX_PPS_MODE

#ifdef RX_PPS_MODE
#define DISABLE_SEND
#define DISABLE_READ
#define DISABLE_WRITE
#endif

#include "utils.c"
#include "ring_buffer.c"

// Update period for packets per second statistic counter
#define PPS_UPDATE_PERIOD INT64_C(1000000000)
#define STAT_INT_LEN "10"

#define N_SAMPLES (32)
#define MIN_PACKET_SIZE 64
#define MAX_CHANNELS 4
#define ETHERNET_HEADER 14
#define ECPRI_COMMON_HEADER 4
#define ECPRI_IQ_HEADER (ECPRI_COMMON_HEADER + 4)
#define ORAN_HEADER 8
#define PACKET_HEADER (ETHERNET_HEADER + ECPRI_IQ_HEADER + ORAN_HEADER)
#define LONG_IQ_PAYLOAD 4416
#define NORMAL_IQ_PAYLOAD 4384
#define MAX_PACKET_SIZE (PACKET_HEADER + LONG_IQ_PAYLOAD)
#define MAX_RX_BURST 1024
#define MAX_TX_BURST 1024

/* eCPRI frame structure

   Ethernet header: 14 bytes
   6         6          2
   +---------+---------+-----------+
   | MAC dst | MAC src | Ethertype |
   +---------+---------+-----------+

   eCPRI common header: 4 bytes

   4b        3b        1b        1         2     
   +---------+----------+--------+---------+---------+
   | Version | Reserved | Concat | Message | Payload |
   |         |          |        |  Type   |  Size   |
   +---------+----------+--------+---------+---------+

   IQ data transfer message header: 4 bytes

   2        2          
   +-------+--------+------------+
   | PC_ID | SEQ_ID | IQ Samples |
   +-------+--------+------------+

   eCPRI concatenated frames

   4                             3          4
   +-----------+------------------+---------+-----------+------------------+
   |   eCPRI   |                  |         |   eCPRI   |                  |
   |  Common   | eCPRI Payload #0 | Padding |  Common   | eCPRI Payload #1 |
   | Header #0 |                  |         | Header #1 |                  |
   +-----------+------------------+---------+-----------+------------------+

   ORAN HEADER:

   Symbol ID:   0 - 6
   Slot ID:     0 - 1
   Subframe ID: 0 - 9
   Frame ID:    0 - 255
   |               8 bits - Frame ID          |
   | 4 bits - Subframe ID | 4 bits - Slot ID  |
   | 2 bits - Slot ID | 6 bits - Symbol ID    |


   IQ sample pair: 32 bit (4 bytes)
   IQ / second / antenna: 30 720 000
   IQ payload per frame: 8832 bytes = 1104 IQ samples
   IQ throughput (4 * 20 MHz): 3 932 160 000 bytes

   1 radio frame = 10 subframes
   1 subframe = 2 slots
   1 slot = 7 symbols

   10MHz:
   Sample Rate: 15.36 MHz
   IQ / packet ~= 4416 / 4 = 1104 
   PPS / channel = 15.36M / 1104 = 13913 pps
   IQ Throughput / channel = 491 Mbps
   Total ethernet throughput = 1.979430336 Gbps

   Comba RU sends 56155.16333333333 PPS

   14                   4                  4               8          8832
   +-----------------+---------------------+----------------+-------------+------------+
   | Ethernet header | eCPRI common header | IQ data header | oRAN header | IQ payload |
   +-----------------+---------------------+----------------+-------------+------------+

 */

typedef struct {
  int64_t counter;
  int64_t freq_counter;
  int64_t freq_ts;
  int64_t freq;
} counter_stat_t;

typedef struct {
  const char * rrh_mac;
  const char * bbu_mac;
  const char * bbu_if;
  const char * log_directory;
  int recv_affinity;
  int recv2_affinity;
  int recv3_affinity;
  int recv4_affinity;
  int send_affinity;
  int statistic_affinity;

  int flow_id;

  int trxw_buf_size;
  int trxr_buf_size;
  int txrx_buf_size;
  int rx_burst;
  int tx_burst;

  int rx_n_channel;
  int tx_n_channel;
  int statistics_refresh_rate_ms;
  int sample_rate;
  int rx_drop_pcm;
  int align_rx;
  int iq_file_trx;
  int disable_rx;
  uint8_t * rx_buf;
  ring_buffer_t * trxr_rbuf[MAX_CHANNELS];
  counter_stat_t * rx_packet_counter; // packets received from RRH
} TRXEcpriState;

typedef struct {
  uint8_t  mac_dst[6];
  uint8_t  mac_src[6];
  uint16_t ether_type;
  uint8_t  ecpri_version;
  uint8_t  ecpri_type;
  uint16_t payload_size;
  uint16_t antenna_id;
  uint8_t seq_id;
  uint8_t seq_id_fixed;
} ecpri_header;

typedef struct {
  uint8_t  mac_dst[6];
  uint8_t  mac_src[6];
  uint16_t ether_type;
  uint8_t  ecpri_version;
  uint8_t  ecpri_type;
  uint16_t payload_size;
  uint16_t antenna_id;
  uint8_t seq_id;
  uint8_t seq_id_fixed;
  uint8_t  oran_header[ORAN_HEADER];
  uint8_t  iq_samples[LONG_IQ_PAYLOAD];
} ecpri_iq_packet;

typedef struct {
  uint8_t  mac_dst[6];
  uint8_t  mac_src[6];
  uint16_t ether_type;
  uint8_t  ecpri_version;
  uint8_t  ecpri_type;
  uint16_t payload_size;
  uint16_t antenna_id;
  uint8_t seq_id;
  uint8_t seq_id_fixed;
  uint8_t  oran_header[ORAN_HEADER];
  uint8_t  gps_time[10];
} ecpri_timing_packet;

static TRXEcpriState s_recv;
static TRXEcpriState s_recv2;
static TRXEcpriState s_recv3;
static TRXEcpriState s_recv4;

// Buffers
static ring_buffer_t trxr_rbuf[MAX_CHANNELS]; // Decoded IQ samples
static ring_buffer_t trxr2_rbuf[MAX_CHANNELS]; // Decoded IQ samples
static ring_buffer_t trxr3_rbuf[MAX_CHANNELS]; // Decoded IQ samples
static ring_buffer_t trxr4_rbuf[MAX_CHANNELS]; // Decoded IQ samples
static ring_buffer_t trxw_rbuf[MAX_CHANNELS]; // Uncompressed IQ samples
static uint8_t rx_buf[MAX_PACKET_SIZE * MAX_RX_BURST];
static uint8_t rx2_buf[MAX_PACKET_SIZE * MAX_RX_BURST];
static uint8_t rx3_buf[MAX_PACKET_SIZE * MAX_RX_BURST];
static uint8_t rx4_buf[MAX_PACKET_SIZE * MAX_RX_BURST];
static uint8_t tx_buf[MAX_PACKET_SIZE * MAX_TX_BURST];

#define TAB_SIZE 65536
float *trx_read_float_tab;
float *trx_write_float_tab;
int16_t *trx_read_int_tab;
int16_t *trx_write_int_tab;

// Counters
static counter_stat_t recv_counter[MAX_CHANNELS]; // IQs received from RRH
static counter_stat_t read_counter; // IQs passed to amarisoft stack
static counter_stat_t write_counter; // IQs to write from TRX
static counter_stat_t sent_counter; // IQs sent to RRH
static counter_stat_t rx_drop_counter; // rx packets dropped by driver
static counter_stat_t tx_drop_counter; // tx packets dropped by driver
static counter_stat_t lost_rx_counter[MAX_CHANNELS]; // gaps between seq_id's
static counter_stat_t rx_packet_counter; // packets received from RRH
static counter_stat_t rx2_packet_counter; // packets received from RRH
static counter_stat_t rx3_packet_counter; // packets received from RRH
static counter_stat_t rx4_packet_counter; // packets received from RRH
static counter_stat_t tx_packet_counter; // packets sent from TRX

// Network
static int send_sockfd;
static int recv_sockfd;
static struct sockaddr_ll connect_sk_addr;
static struct sockaddr_ll recv_connect_sk_addr;
static ecpri_iq_packet standard_iq_packet;

#define MAX_IQ_FILE_SIZE 1228800
static int iq_file_size;
static uint8_t iq_file_bytes[MAX_IQ_FILE_SIZE];
static int iq_file_index = 0;

static void print_debug(FILE * f, int print_header) {
  char buffer[200];
  if(print_header) {
    sprintf(buffer, 
        "        "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "%" STAT_INT_LEN "s "
        "\n",
        "TRXR RA",
        "TRXW RA",
        "TRXR WA",
        "TRXW WA",
        "TRXW RI",
        "TRXW WI",
        "TRXR RI",
        "TRXR WI");
    for(int i = 0; buffer[i] != '\0'; i++)
      if(buffer[i] == ' ')
        buffer[i] = '-';
    fprintf(f, "%s", buffer);
  }
  sprintf(buffer, 
      "        "
      "%" STAT_INT_LEN "d "
      "%" STAT_INT_LEN "d "
      "%" STAT_INT_LEN "d "
      "%" STAT_INT_LEN "d "
      "%" STAT_INT_LEN "d "
      "%" STAT_INT_LEN "d "
      "%" STAT_INT_LEN "d "
      "%" STAT_INT_LEN "d "
      "\n",
      rbuf_read_amount(&trxr_rbuf[0]),
      rbuf_read_amount(&trxw_rbuf[0]),
      rbuf_write_amount(&trxr_rbuf[0]),
      rbuf_write_amount(&trxw_rbuf[0]),
      trxw_rbuf[0].read_index,
      trxw_rbuf[0].write_index,
      trxr_rbuf[0].read_index,
      trxr_rbuf[0].write_index);
  for(int i = 0; buffer[i] != '\0'; i++)
    if(buffer[i] == ' ')
      buffer[i] = '-';
  fprintf(f, "%s", buffer);
}

int pow10i(int n) {
  int r = 1;
  for(int i = 0; i < n; i++)
    r *= 10;
  return r;
}
int n_digits(int64_t n) {
    int r = 1;
    while (n > 9) {
        n /= 10;
        r++;
    }
    return r;
}
static void int64_to_6s(char * s, int64_t x) {
  double n = n_digits((double) x);
  if(n < 7)
    sprintf(s, "%6d", (int) x);
  else if(n < 10) {
    int a = x / 1000000;
    int b = (x % 1000000) / pow10i(n - 4);
    sprintf(s, "%d.%dM", a, b);
  }
  else if(n < 13) {
    int a = x / 1000000000;
    int b = (x % 1000000000) / pow10i(n - 4);
    sprintf(s, "%d.%dG", a, b);
  }
  else if(n < 16) {
    int c = x / 1000000;
    int a = c / 1000000;
    int b = ((c / 1000) % 1000) / pow10i(n - 13);
    sprintf(s, "%d.%dT", a, b);
  }
  s[6] = '\0';
}

static void print_stats(FILE * f, int print_header) {
  char s[300];
  int offset = 0;
  if(print_header) {
    fprintf(f, 
        "%6s %6s %6s %6s %6s %6s %6s %6s %6s %6s %6s %6s %6s  %6s %6s %6s %6s  %6s %6s %6s %6s %6s  %6s %6s %6s %6s %6s\n",
        "rxdrop",
        "txdrop",
        "lstrx0",
        "lstrx1",
        "lstrx2",
        "lstrx3",
        "recv 0",
        "recv 1",
        "recv 2",
        "recv 3",
        "read",
        "write",
        "sent",
        "recv f",
        "read f",
        "wrte f",
        "sent f",
        "rx pkt",
        "rx2pkt",
        "tx pkt",
        "rx pps",
        "tx pps",
        "rxbuf0",
        "rxbuf1",
        "rxbuf2",
        "rxbuf3",
        "txbufw");
  }
  offset = 0;
  int64_to_6s(s + 7 * offset++, rx_drop_counter.counter);
  int64_to_6s(s + 7 * offset++, tx_drop_counter.counter);
  int64_to_6s(s + 7 * offset++, lost_rx_counter[0].counter);
  int64_to_6s(s + 7 * offset++, lost_rx_counter[1].counter);
  int64_to_6s(s + 7 * offset++, lost_rx_counter[2].counter);
  int64_to_6s(s + 7 * offset++, lost_rx_counter[3].counter);
  int64_to_6s(s + 7 * offset++, recv_counter[0].counter);
  int64_to_6s(s + 7 * offset++, recv_counter[1].counter);
  int64_to_6s(s + 7 * offset++, recv_counter[2].counter);
  int64_to_6s(s + 7 * offset++, recv_counter[3].counter);
  int64_to_6s(s + 7 * offset++, read_counter.counter);
  int64_to_6s(s + 7 * offset++, write_counter.counter);
  int64_to_6s(s + 7 * offset++, sent_counter.counter);
  int64_to_6s(s + 7 * offset++, recv_counter[0].freq +
                                recv_counter[1].freq +
                                recv_counter[2].freq +
                                recv_counter[3].freq);
  int64_to_6s(s + 7 * offset++, read_counter.freq);
  int64_to_6s(s + 7 * offset++, write_counter.freq);
  int64_to_6s(s + 7 * offset++, sent_counter.freq);
  int64_to_6s(s + 7 * offset++, rx_packet_counter.counter);
  int64_to_6s(s + 7 * offset++, rx2_packet_counter.counter + rx3_packet_counter.counter + rx4_packet_counter.counter);
  int64_to_6s(s + 7 * offset++, tx_packet_counter.counter);
  int64_to_6s(s + 7 * offset++, rx_packet_counter.freq + rx2_packet_counter.freq + rx3_packet_counter.freq + rx4_packet_counter.freq);
  int64_to_6s(s + 7 * offset++, tx_packet_counter.freq);
  int64_to_6s(s + 7 * offset++, rbuf_write_amount(&trxr_rbuf[0]) / 4);
  int64_to_6s(s + 7 * offset++, rbuf_write_amount(&trxr_rbuf[1]) / 4);
  int64_to_6s(s + 7 * offset++, rbuf_write_amount(&trxr_rbuf[2]) / 4);
  int64_to_6s(s + 7 * offset++, rbuf_write_amount(&trxr_rbuf[3]) / 4);
  int64_to_6s(s + 7 * offset++, rbuf_write_amount(&trxw_rbuf[0]) / 4);
  offset = 0;
  fprintf(f, 
    "%6s %6s %6s %6s %6s %6s %6s %6s %6s %6s %6s %6s %6s  %6s %6s %6s %6s  %6s %6s %6s %6s %6s  %6s %6s %6s %6s %6s\n",
    s + 7 * 0,
    s + 7 * 1,
    s + 7 * 2,
    s + 7 * 3,
    s + 7 * 4,
    s + 7 * 5,
    s + 7 * 6,
    s + 7 * 7,
    s + 7 * 8,
    s + 7 * 9,
    s + 7 * 10,
    s + 7 * 11,
    s + 7 * 12,
    s + 7 * 13,
    s + 7 * 14,
    s + 7 * 15,
    s + 7 * 16,
    s + 7 * 17,
    s + 7 * 18,
    s + 7 * 19,
    s + 7 * 20,
    s + 7 * 21,
    s + 7 * 22,
    s + 7 * 23,
    s + 7 * 24,
    s + 7 * 25,
    s + 7 * 26);
}

static void log_exit(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
  char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " EXIT [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  fprintf(stderr, "%s\n", line);

  // Dump useful information
  print_stats(stderr, 1);
  print_debug(stderr, 1);
  fprintf(stderr, "TRXW RBUF: ri %d wi %d ra %d wa %d\n",
      trxw_rbuf[0].read_index, trxw_rbuf[0].write_index, rbuf_read_amount(&trxw_rbuf[0]),
      rbuf_write_amount(&trxw_rbuf[0]));
  fprintf(stderr, "TRXR RBUF: ri %d wi %d ra %d wa %d\n",
      trxr_rbuf[0].read_index, trxr_rbuf[0].write_index, rbuf_read_amount(&trxr_rbuf[0]),
      rbuf_write_amount(&trxr_rbuf[0]));

  fflush(stdout);
  fflush(stderr);
  exit(EXIT_FAILURE);
}

typedef struct {
  uint8_t frame_id;
  uint8_t subframe_id;
  uint8_t slot_id;
  uint8_t symbol_id;
} oran_counters_t;

void increment_oran_counters(oran_counters_t * o) {
  o->symbol_id = (o->symbol_id + 1) % 7;
  if(!o->symbol_id) {
    o->slot_id = (o->slot_id + 1) % 2;
    if(!o->slot_id) {
      o->subframe_id = (o->subframe_id + 1) % 10;
      if(!o->subframe_id) {
        o->frame_id = (o->frame_id + 1) % 256;
      }
    }
  }
}

int int_oran_counters(oran_counters_t * o) {
  return o->symbol_id + o->slot_id * 7 + o->subframe_id * 2 * 7 + o->frame_id * 2 * 7 * 10;
}
int diff_oran_counters(oran_counters_t * o1, oran_counters_t * o2) {
  int n = (6 + 1 * 7 + 9 * 2 * 7 + 255 * 2 * 7 * 10 + 1);
  return (int_oran_counters(o1) + n - int_oran_counters(o2)) % n;
}


void decode_oran_header(oran_counters_t * o, uint8_t * data) {
  o->frame_id    =  data[1];
  o->subframe_id = (data[2] >> 4) & ((1 << 4) - 1);
  o->slot_id     = (uint8_t) (*((uint16_t *) (data + 3)) >> 6) & ((1 << 6) - 1);
  o->symbol_id   = (data[3]     ) & ((1 << 6) - 1);
}
void encode_oran_header(oran_counters_t * o, uint8_t * data) {
  data[1] = o->frame_id;
  data[2] = o->subframe_id << 4;
  data[3] = o->symbol_id;
  *((uint16_t *) (data + 3)) |= o->slot_id << 6 ;
}

static void init_counter(counter_stat_t * c) {
  c->counter = 0;
  c->freq_counter = 0;
  c->freq_ts = 0;
  c->freq = 0;
}
static void update_counter_freq(counter_stat_t * c) {
  struct timespec _ts;
  int64_t ts;
  clock_gettime(CLOCK_TAI, &_ts);
  ts = ts_to_int(_ts);
  if((ts - c->freq_ts) > PPS_UPDATE_PERIOD) {
    if(c->freq_ts)
      c->freq = ((c->counter - c->freq_counter) * NSEC_PER_SEC) / (ts - c->freq_ts);
    c->freq_counter = c->counter;
    c->freq_ts = ts;
  }
}
static void update_counter(counter_stat_t * c, int64_t v) {
  c->counter += v;
}

void print_bytes(const char * str, uint8_t * data, int count) {
  printf("%s", str);
  for(int i = 0; i < count; i++)
    printf("%02x", data[i]);
  printf(", ");
}
void print_packet(ecpri_iq_packet * iq_packet) {
  oran_counters_t o;
  print_bytes("mac_dst: "      , (uint8_t *) &iq_packet->mac_dst,       6);
  print_bytes("mac_src: "      , (uint8_t *) &iq_packet->mac_src,       6);  
  print_bytes("ethertype: "    , (uint8_t *) &iq_packet->ether_type,    2);
  print_bytes("ecpri_version: ", (uint8_t *) &iq_packet->ecpri_version, 1);
  print_bytes("ecpri_type: "   , (uint8_t *) &iq_packet->ecpri_type,    1);
  print_bytes("payload_size: " , (uint8_t *) &iq_packet->payload_size,  2);
  print_bytes("antenna_id: "   , (uint8_t *) &iq_packet->antenna_id,    2);
  print_bytes("seq_id: "       , (uint8_t *) &iq_packet->seq_id,        1);
  print_bytes("seq_id_fixed: " , (uint8_t *) &iq_packet->seq_id_fixed,  1);
  printf("\n");
  decode_oran_header(&o, iq_packet->oran_header);
  printf("symbol = %d , slot = %d, subframe = %d, frame = %d\n",
         o.symbol_id,
         o.slot_id,
         o.subframe_id,
         o.frame_id);
}

static void *recv_thread(void *p) {
#ifdef DISABLE_RECV
  pthread_exit(EXIT_SUCCESS);
#endif

  cpu_set_t mask;
  int stop = 0;

  uint16_t antenna_id;
  uint16_t payload_size;
  oran_counters_t o;
  oran_counters_t prev_o[MAX_CHANNELS];
  int prev_o_defined[MAX_CHANNELS];
  int lost_pkts = 0;
  
  int prev_lost_pkts = 0;
  int received_hyper_frames = 0;
  int ready = 0;

  struct mmsghdr msgh[MAX_RX_BURST];
  struct iovec msgv[MAX_RX_BURST];
  ecpri_header * header;
  ecpri_iq_packet * iq_packet;
  ecpri_timing_packet * timing_packet;
  TRXEcpriState * s = (TRXEcpriState *) p;

  log_info("RECV_THREAD", "Thread init (Affinity %d)", s->recv_affinity);
  // Set thread CPU affinity
  CPU_ZERO(&mask);
  CPU_SET(s->recv_affinity, &mask);
  if (sched_setaffinity(0, sizeof(mask), &mask))
    error(EXIT_FAILURE, errno,
        "Could not set CPU affinity to CPU %d\n", s->recv_affinity);

  memset((uint8_t *) prev_o, 0, sizeof(oran_counters_t) * MAX_CHANNELS);
  memset((uint8_t *) prev_o_defined, 0, sizeof(int) * MAX_CHANNELS);

  for(int64_t i = 0;; i++) {
  
    if(s->disable_rx) {
      usleep(10000);
      continue;
    }

    // Reset data structures for recv messages 
    memset(msgv, 0, sizeof(msgv));
    memset(msgh, 0, sizeof(msgh));
    for(int j = 0; j < s->rx_burst; j++) {
      msgv[j].iov_base = s->rx_buf + j * MAX_PACKET_SIZE;
      msgv[j].iov_len = MAX_PACKET_SIZE;
      msgh[j].msg_hdr.msg_iov = &msgv[j];
      msgh[j].msg_hdr.msg_iovlen = 1;
    }

    // Receive at most rx_burst messages
    int ret = recvmmsg(recv_sockfd, msgh, s->rx_burst, 0, NULL);
    if(ret <= -1)
      error(EXIT_FAILURE, errno, "recvmmsg error");
#ifdef RX_PPS_MODE
    update_counter(s->rx_packet_counter, ret);
    continue;
#endif
    if((i % 100000) < s->rx_drop_pcm)
      continue;
    // Process each received message
    for(int j = 0; j < ret; j++) {
      header = (ecpri_header*) (s->rx_buf + j * MAX_PACKET_SIZE);

      // Discard packet if it is not eCPRI
      if(header->ether_type != 0xfeae)
        continue;
      // Stop if packet has unexpected size
      if((msgh + j)->msg_len < (PACKET_HEADER + 32)) {
        stop = 1;
        log_info("RECV_THREAD", "Packet doesn't have correct size (%d < %d)", (msgh + j)->msg_len, PACKET_HEADER + 32);
        break;
      }
      // If packet is a timing packet
      if(header->ecpri_type == 255) {
        timing_packet = (ecpri_timing_packet*) (s->rx_buf + j * MAX_PACKET_SIZE);
        log_info("RECV_THREAD", "GPS TIME: %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
                 timing_packet->gps_time[0],
                 timing_packet->gps_time[1],
                 timing_packet->gps_time[2],
                 timing_packet->gps_time[3],
                 timing_packet->gps_time[4],
                 timing_packet->gps_time[5],
                 timing_packet->gps_time[6],
                 timing_packet->gps_time[7],
                 timing_packet->gps_time[8],
                 timing_packet->gps_time[9]);
        continue;
      }
      // Exit if packet is neither timing nor IQ
      else if(header->ecpri_type != 0) {
        log_info("RECV_THREAD", "Unknown eCPRI type: %d\n", header->ecpri_type);
        stop = 1; break;
      }
      antenna_id = ntohs(header->antenna_id);

      // Exit if antenna ID is not in range
      if(antenna_id > s->rx_n_channel || antenna_id < 0) {
        log_info("RECV_THREAD", "Wrong Antenna ID: %d\n", antenna_id);
        stop = 1; break;
      }
      // Exit if there is no more space in the buffer
      if(rbuf_write_amount(s->trxr_rbuf[antenna_id]) <= s->trxr_rbuf[antenna_id]->block_len) {
        //rbuf_increment_read(s->trxr_rbuf[antenna_id], s->trxr_rbuf[antenna_id]->block_len);
        //update_counter(&rx_drop_counter, 1);
        log_exit("RECV_THREAD", "No more space in %s buffer (%ld < %ld)",
                 s->trxr_rbuf[antenna_id]->name,
                 rbuf_write_amount(s->trxr_rbuf[antenna_id]),
                 s->trxr_rbuf[antenna_id]->block_len);
      }

      iq_packet = (ecpri_iq_packet*) (s->rx_buf + j * MAX_PACKET_SIZE);

      // ORAN counters
      decode_oran_header(&o, iq_packet->oran_header);

      lost_pkts = 0;
      if(prev_o_defined[antenna_id]) {
        lost_pkts = diff_oran_counters(&o, &prev_o[antenna_id]) - 1;
        if(lost_pkts < 0) {
          printf("o = %d %d %d %d\n", o.symbol_id, o.slot_id, o.subframe_id, o.frame_id);
          printf("prev_o = %d %d %d %d\n", prev_o[antenna_id].symbol_id,
                                           prev_o[antenna_id].slot_id,
                                           prev_o[antenna_id].subframe_id,
                                           prev_o[antenna_id].frame_id);
          printf("lost_pkts = %d\n", lost_pkts);
          stop = 1 ; break ;
        }
      }

      if(s->align_rx) {
        if((o.symbol_id   == 0) &&
           (o.slot_id     == 0) &&
           (o.subframe_id == 0) &&
           (o.frame_id    == 0)) {
          received_hyper_frames++;
          // If we received at least 2 hyperframes and no packet has been lost
          if((received_hyper_frames > 1) && (prev_lost_pkts == lost_pkts)) {
            ready = 1;
          } else if(received_hyper_frames > 1) {
            printf("%d != %d\n", prev_lost_pkts, lost_pkts);
          }
          prev_lost_pkts = lost_pkts;
        }
        if(!ready)
          continue;
      }

      if(prev_o_defined[antenna_id]) {
        update_counter(&lost_rx_counter[antenna_id], lost_pkts);
      }

      for(int k = 0; k < lost_pkts; k++) {
        if(prev_o[antenna_id].symbol_id == 0)
          payload_size = LONG_IQ_PAYLOAD;
        else
          payload_size = NORMAL_IQ_PAYLOAD;
        memset(rbuf_write(s->trxr_rbuf[antenna_id]), 0, payload_size);
        rbuf_increment_write(s->trxr_rbuf[antenna_id], payload_size);
        increment_oran_counters(&prev_o[antenna_id]);
      }
      payload_size = ntohs(iq_packet->payload_size) - 12;

      //int mismatch = 0;
      //for(int k = 0; k < payload_size / 2; k++) {
      //  if(iq_packet->iq_samples[2 * k] != 0x62 || iq_packet->iq_samples[2 * k +1] != 0x8f) {
      //    printf("2 * k = %d (%d)\n", 2 * k, payload_size);
      //    printf("%x%x \n", iq_packet->iq_samples[2 * k], iq_packet->iq_samples[2 * k+1]);
      //    print_packet(iq_packet);
      //    mismatch++;
      //    break;
      //  }
      //}
      //if(mismatch) {
      //  //printf("\n%d/%d mismatch\n", mismatch, payload_size/2);
      //  for(int k = 0; k < payload_size / 2; k++) {
      //    printf("%x%x ", iq_packet->iq_samples[2 * k], iq_packet->iq_samples[2 * k+1]);
      //  }
      //  printf("\n");
      //}

      if((o.symbol_id == 0 && payload_size != LONG_IQ_PAYLOAD) ||
         (o.symbol_id != 0 && payload_size != NORMAL_IQ_PAYLOAD)) {
        log_info("RECV_THREAD",
                 "Unexpected payload size (symbol is %d but payload size is %d)",
                 o.symbol_id, payload_size);
        stop = 1; break;
      }

      memcpy(rbuf_write(s->trxr_rbuf[antenna_id]), iq_packet->iq_samples, payload_size);
      rbuf_increment_write(s->trxr_rbuf[antenna_id], payload_size);

      update_counter(&recv_counter[antenna_id], payload_size / 4);
      update_counter(s->rx_packet_counter, 1);

      prev_o[antenna_id] = o;
      prev_o_defined[antenna_id] = 1;
    }
    if(stop) {
      for(int j = 0; j < ret; j++) {
        iq_packet = (ecpri_iq_packet*) (s->rx_buf + j * MAX_PACKET_SIZE);
        print_packet(iq_packet);
      }
      log_exit("RECV_THREAD", "Exiting");
    }
  }
  pthread_exit(EXIT_SUCCESS);
}

// Send as soon as packets are encoded
static void *send_thread(void *p) {
#ifdef DISABLE_SEND
  pthread_exit(EXIT_SUCCESS);
#endif

  cpu_set_t mask;
  struct timespec initial;
  struct mmsghdr msgh[MAX_TX_BURST];
  struct iovec msgv[MAX_TX_BURST];
  int tx_seq_id = 0;
  int iq_sent;
  int tx_buf_index = 0;
  int to_read;
  int payload_size = 0;
  oran_counters_t o;
  ecpri_iq_packet * iq_packet;
  TRXEcpriState * s = (TRXEcpriState *) p;

  log_info("SEND_THREAD", "Thread init");
  // Set thread CPU affinity
  CPU_ZERO(&mask);
  CPU_SET(s->send_affinity, &mask);
  if (sched_setaffinity(0, sizeof(mask), &mask))
    error(EXIT_FAILURE, errno,
        "Could not set CPU affinity to CPU %d\n", s->send_affinity);

  memset(msgv, 0, sizeof(msgv));
  memset(msgh, 0, sizeof(msgh));
  for(int j = 0; j < s->tx_burst; j++) {
    msgh[j].msg_hdr.msg_name = &connect_sk_addr;
    msgh[j].msg_hdr.msg_namelen = sizeof(connect_sk_addr);
    msgh[j].msg_hdr.msg_iov = &msgv[j];
    msgh[j].msg_hdr.msg_iovlen = 1;
  }
  clock_gettime(CLOCK_TAI, &initial);
  log_info("SEND_THREAD", "Starting loop");

  memset((uint8_t*) &o, 0, sizeof(oran_counters_t));

  for(int64_t i = 1;; i++) {
    // Send at most tx_burst packets
    int to_send = s->tx_burst / s->tx_n_channel;
    for(int k = 0; k < s->tx_n_channel; k++) {
      to_read = rbuf_read_amount(&trxw_rbuf[k]) / LONG_IQ_PAYLOAD;
      if(to_read < to_send)
        to_send = to_read;
    }
    to_send *= s->tx_n_channel;

    iq_sent = 0;
    tx_buf_index = 0;
    for(int encoded = 0; encoded < to_send;) {
      for(uint16_t antenna_id = 0 ; antenna_id < s->tx_n_channel; antenna_id++) {
        iq_packet = (ecpri_iq_packet*) (tx_buf + tx_buf_index);
        memcpy((uint8_t *) iq_packet, (uint8_t*) &standard_iq_packet, PACKET_HEADER);

        iq_packet->antenna_id = htons(antenna_id);
        iq_packet->seq_id = (uint8_t) tx_seq_id;
        if(o.symbol_id == 0)
          payload_size = LONG_IQ_PAYLOAD;
        else
          payload_size = NORMAL_IQ_PAYLOAD;
        iq_packet->payload_size  = htons((uint16_t) (payload_size + 12));
        msgv[encoded].iov_base = (uint8_t *) iq_packet;
        msgv[encoded].iov_len = payload_size + PACKET_HEADER;
        tx_seq_id = (tx_seq_id + 1) % 256;
        encode_oran_header(&o, iq_packet->oran_header);

        to_read = rbuf_read_amount(&trxw_rbuf[antenna_id]);


        memcpy(iq_packet->iq_samples, rbuf_read(&trxw_rbuf[antenna_id]), payload_size);
        //int mismatch = 0;
        //for(int k = 0; k < payload_size / 2; k++) {
        //  if(iq_packet->iq_samples[2 * k] != 0x62 || iq_packet->iq_samples[2 * k +1] != 0x8f) {
        //    printf("SEND 2 * k = %d (%d)\n", 2 * k, payload_size);
        //    printf("SEND %x%x \n", iq_packet->iq_samples[2 * k], iq_packet->iq_samples[2 * k+1]);
        //    printf("SEND %x%x \n", rbuf_read(&trxw_rbuf[antenna_id])[2 * k], rbuf_read(&trxw_rbuf[antenna_id])[2 * k+1]);
        //    printf("SEND to_read = %d\n", to_read);
        //    print_packet(iq_packet);
        //    mismatch++;
        //    break;
        //  }
        //}
        //if(mismatch) {
        //  //printf("\n%d/%d mismatch\n", mismatch, payload_size/2);
        //  for(int k = 0; k < payload_size / 2; k++) {
        //    printf("%x%x ", iq_packet->iq_samples[2 * k], iq_packet->iq_samples[2 * k+1]);
        //  }
        //  printf("\n");
        //}


        rbuf_increment_read(&trxw_rbuf[antenna_id], payload_size);
        encoded++;
        iq_sent += (payload_size) / 4;
        tx_buf_index += payload_size + PACKET_HEADER;
      }
      increment_oran_counters(&o);
    }
    usleep(1); // TODO

    for(int j = 0; j < to_send;) {

      int ret = sendmmsg(send_sockfd, msgh + j, (to_send - j), 0);
      if(ret <= 0)
        error(EXIT_FAILURE, errno, "sendmmsg error (returned %d)", ret);
      j += ret;
      update_counter(&tx_packet_counter, ret);
    }
    update_counter(&sent_counter, iq_sent);
  }
  pthread_exit(EXIT_SUCCESS);
}

static void *statistic_thread(void *p) {
#ifdef DISABLE_STAT
  pthread_exit(EXIT_SUCCESS);
#endif
  struct timespec next, initial;
  cpu_set_t mask;
  TRXEcpriState * s = (TRXEcpriState *) p;
  FILE * stats_file_desc;

  log_info("STATISTIC_THREAD", "Thread init");

  char stats_file_name[256];
  memset(stats_file_name, '\0', 256);
  sprintf(stats_file_name, "%s/ecpri.stats", s->log_directory);
  stats_file_desc = fopen(stats_file_name, "w+");

  if(!stats_file_desc)
    error(EXIT_FAILURE, errno, "Couldn't open %s\n", stats_file_name);

  for(int i = 0; i < 10; i++)
    fprintf(stats_file_desc, "\n");

  // Set thread CPU affinity
  CPU_ZERO(&mask);
  CPU_SET(s->statistic_affinity, &mask);
  if (sched_setaffinity(0, sizeof(mask), &mask))
    error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n", s->statistic_affinity);

  clock_gettime(CLOCK_TAI, &initial);
  next = initial;
  for(int64_t i = 0;; i++) {
    add_ns(&next, ((int64_t) s->statistics_refresh_rate_ms) * 1000000L);
    print_stats(stats_file_desc, (i % 17) == 0);
#ifdef DEBUG
    //print_debug(stats_file_desc, ((i + 9) % 17) == 0);
#endif
    fflush(stats_file_desc);

    update_counter_freq(&rx_drop_counter);
    update_counter_freq(&tx_drop_counter);
    for(int i = 0; i < MAX_CHANNELS; i++)
      update_counter_freq(&recv_counter[i]);
    update_counter_freq(&read_counter);
    update_counter_freq(&write_counter);
    update_counter_freq(&sent_counter);
    update_counter_freq(&rx_packet_counter);
    update_counter_freq(&rx2_packet_counter);
    update_counter_freq(&rx3_packet_counter);
    update_counter_freq(&rx4_packet_counter);
    update_counter_freq(&tx_packet_counter);

    clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &next, NULL);
  }
  pthread_exit(EXIT_SUCCESS);
}

static int start_threads(TRXEcpriState * s) {
  pthread_t recv_pthread;
  pthread_t recv2_pthread;
  pthread_t recv3_pthread;
  pthread_t recv4_pthread;
  pthread_t send_pthread;
  pthread_t statistic_pthread;
  struct sched_param recv_param;
  struct sched_param recv2_param;
  struct sched_param recv3_param;
  struct sched_param recv4_param;
  struct sched_param send_param;
  struct sched_param statistic_param;
  pthread_attr_t recv_attr;
  pthread_attr_t recv2_attr;
  pthread_attr_t recv3_attr;
  pthread_attr_t recv4_attr;
  pthread_attr_t send_attr;
  pthread_attr_t statistic_attr;

  memcpy((uint8_t*) &s_recv,  s, sizeof(TRXEcpriState));
  memcpy((uint8_t*) &s_recv2, s, sizeof(TRXEcpriState));
  memcpy((uint8_t*) &s_recv3, s, sizeof(TRXEcpriState));
  memcpy((uint8_t*) &s_recv4, s, sizeof(TRXEcpriState));
  s_recv.recv_affinity  = s->recv_affinity;
  s_recv2.recv_affinity = s->recv2_affinity;
  s_recv3.recv_affinity = s->recv3_affinity;
  s_recv4.recv_affinity = s->recv4_affinity;
  s_recv.rx_buf  = rx_buf;
  s_recv2.rx_buf = rx2_buf;
  s_recv3.rx_buf = rx3_buf;
  s_recv4.rx_buf = rx4_buf;
  for(int i = 0; i < MAX_CHANNELS; i++) {
    s_recv.trxr_rbuf[i]  = &trxr_rbuf[i];
    s_recv2.trxr_rbuf[i] = &trxr2_rbuf[i];
    s_recv3.trxr_rbuf[i] = &trxr3_rbuf[i];
    s_recv4.trxr_rbuf[i] = &trxr4_rbuf[i];
  }
  s_recv.rx_packet_counter  = &rx_packet_counter;
  s_recv2.rx_packet_counter = &rx2_packet_counter;
  s_recv3.rx_packet_counter = &rx3_packet_counter;
  s_recv4.rx_packet_counter = &rx4_packet_counter;

  log_info("TRX_ECPRI", "Starting threads");

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&recv_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  // Set a specific stack size
  if (pthread_attr_setstacksize(&recv_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&recv_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  recv_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&recv_attr, &recv_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  /* Use scheduling parameters of attr */
  if (pthread_attr_setinheritsched(&recv_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_attr_init(&recv2_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&recv2_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&recv2_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  recv2_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&recv2_attr, &recv2_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&recv2_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");
  if (pthread_attr_init(&recv3_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&recv3_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&recv3_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  recv3_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&recv3_attr, &recv3_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&recv3_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");
  if (pthread_attr_init(&recv4_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&recv4_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&recv4_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  recv4_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&recv4_attr, &recv4_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&recv4_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_attr_init(&send_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&send_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&send_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  send_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&send_attr, &send_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&send_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_attr_init(&statistic_attr))
    log_error("TRX_ECPRI", "init pthread attributes failed\n");
  if (pthread_attr_setstacksize(&statistic_attr, PTHREAD_STACK_MIN))
    log_error("TRX_ECPRI", "pthread setstacksize failed\n");
  if (pthread_attr_setschedpolicy(&statistic_attr, SCHED_FIFO))
    log_error("TRX_ECPRI", "pthread setschedpolicy failed\n");
  statistic_param.sched_priority = 97;
  if (pthread_attr_setschedparam(&statistic_attr, &statistic_param))
    log_error("TRX_ECPRI", "pthread setschedparam failed\n");
  if (pthread_attr_setinheritsched(&statistic_attr, PTHREAD_EXPLICIT_SCHED))
    log_error("TRX_ECPRI", "pthread setinheritsched failed\n");

  if (pthread_create(&statistic_pthread, NULL, statistic_thread, s))
    error(EXIT_FAILURE, errno, "Couldn't create statistic thread");
  usleep(1000 * 20);
  if (pthread_create(&send_pthread, NULL, send_thread, s))
    error(EXIT_FAILURE, errno, "Couldn't create send thread");
  usleep(1000 * 500);
  if (pthread_create(&recv_pthread, NULL, recv_thread, &s_recv))
    error(EXIT_FAILURE, errno, "Couldn't create recv thread");
  usleep(1000 * 500);
#ifndef DISABLE_RECV2
  if (pthread_create(&recv2_pthread, NULL, recv_thread, &s_recv2))
    error(EXIT_FAILURE, errno, "Couldn't create recv2 thread");
  usleep(1000 * 500);
  if (pthread_create(&recv3_pthread, NULL, recv_thread, &s_recv3))
    error(EXIT_FAILURE, errno, "Couldn't create recv3 thread");
  usleep(1000 * 500);
  if (pthread_create(&recv4_pthread, NULL, recv_thread, &s_recv4))
    error(EXIT_FAILURE, errno, "Couldn't create recv4 thread");
#endif

  return 0;
}

int start(TRXEcpriState * s) {
  uint8_t dst_mac[6];
  uint8_t src_mac[6];
  int if_index;
  FILE *read_ptr;

  log_info("TRX_ECPRI", "raw socket setup");

  /* From the example above: tcpdump -i em1 port 22 -dd */
  struct sock_filter code[] = {
    { 0x28, 0, 0, 0x0000000c },
    { 0x15, 0, 1, 0x0000aefe },
    { 0x06, 0, 0, 0x00040000 },
    { 0x06, 0, 0, 0x00000000 },
  };

  struct sock_fprog bpf = {
          .len = 4,
          .filter = code,
  };

  //set_latency_target();

  posix_memalign((void **)&trx_read_float_tab,  16, sizeof(float)   * TAB_SIZE);
  posix_memalign((void **)&trx_write_float_tab, 16, sizeof(float)   * TAB_SIZE);
  posix_memalign((void **)&trx_read_int_tab,    16, sizeof(int16_t) * TAB_SIZE);
  posix_memalign((void **)&trx_write_int_tab,   16, sizeof(int16_t) * TAB_SIZE);

  init_counter(&rx_drop_counter);
  init_counter(&tx_drop_counter);
  for(int i = 0; i < MAX_CHANNELS; i++) {
    init_counter(&lost_rx_counter[i]);
    init_counter(&recv_counter[i]);
  }
  init_counter(&read_counter);
  init_counter(&write_counter);
  init_counter(&sent_counter);
  init_counter(&rx_packet_counter);
  init_counter(&rx2_packet_counter);
  init_counter(&rx3_packet_counter);
  init_counter(&rx4_packet_counter);
  init_counter(&tx_packet_counter);

  for(int i = 0; i < MAX_CHANNELS; i++) {
    char name[256];
    sprintf(name, "TRXWrite %d", i);
    init_rbuf(&trxw_rbuf[i], name, s->trxw_buf_size, NORMAL_IQ_PAYLOAD * 6 + LONG_IQ_PAYLOAD);
  }
  for(int i = 0; i < MAX_CHANNELS; i++) {
    char name[256];
    sprintf(name, "TRXRead %d", i);
    init_rbuf(&trxr_rbuf[i], name, s->trxr_buf_size, NORMAL_IQ_PAYLOAD * 6 + LONG_IQ_PAYLOAD);
#ifndef DISABLE_RECV2
    sprintf(name, "TRXRead2 %d", i);
    init_rbuf(&trxr2_rbuf[i], name, s->trxr_buf_size, NORMAL_IQ_PAYLOAD * 6 + LONG_IQ_PAYLOAD);
    sprintf(name, "TRXRead3 %d", i);
    init_rbuf(&trxr3_rbuf[i], name, s->trxr_buf_size, NORMAL_IQ_PAYLOAD * 6 + LONG_IQ_PAYLOAD);
    sprintf(name, "TRXRead4 %d", i);
    init_rbuf(&trxr4_rbuf[i], name, s->trxr_buf_size, NORMAL_IQ_PAYLOAD * 6 + LONG_IQ_PAYLOAD);
#else
    sprintf(name, "TRXRead2 %d", i);
    init_rbuf(&trxr2_rbuf[i], name, 1, 1);
    sprintf(name, "TRXRead3 %d", i);
    init_rbuf(&trxr3_rbuf[i], name, 1, 1);
    sprintf(name, "TRXRead4 %d", i);
    init_rbuf(&trxr4_rbuf[i], name, 1, 1);
#endif
  }

  if (!(if_index = if_nametoindex(s->bbu_if))) {
    perror("if_nametoindex");
    return 1;
  }

  if(sscanf(s->rrh_mac, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx%*c",
        &dst_mac[0],
        &dst_mac[1],
        &dst_mac[2],
        &dst_mac[3],
        &dst_mac[4],
        &dst_mac[5]) != 6)
    fprintf(stderr, "Invalid RRH MAC address\n");
  if(sscanf(s->bbu_mac, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx%*c",
        &src_mac[0],
        &src_mac[1],
        &src_mac[2],
        &src_mac[3],
        &src_mac[4],
        &src_mac[5]) != 6)
    fprintf(stderr, "Invalid BBU MAC address\n");

  if ((send_sockfd = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) == -1) {
    perror("Socket Error");
    return 1;
  }
  if ((recv_sockfd = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) == -1) {
    perror("Socket Error");
    return 1;
  }
  const int len = strnlen(s->bbu_if, IFNAMSIZ);
  if (len == IFNAMSIZ) {
    fprintf(stderr, "Too long iface name");
    return 1;
  }
  recv_connect_sk_addr.sll_ifindex = if_index;
  recv_connect_sk_addr.sll_family = AF_PACKET;
  recv_connect_sk_addr.sll_protocol = htons(ETH_P_ALL);
  recv_connect_sk_addr.sll_halen = ETH_ALEN;
  for(int i = 0; i < 6; i++)
    connect_sk_addr.sll_addr[i] = src_mac[i];
  // Only receive packets on configured BBU Ethernet interface
  if((bind(recv_sockfd , (struct sockaddr *) &recv_connect_sk_addr ,
      sizeof(recv_connect_sk_addr))) ==-1)
  {
    perror("bind: ");
    return 1;
  }
  // Only receive incoming packets, don't capture packets we are sending
  if(setsockopt(recv_sockfd, SOL_PACKET, PACKET_IGNORE_OUTGOING, &(int){1}, sizeof(int))) {
    perror("setsockopt PACKET_IGNORE_OUTGOING: ");
    return 1;
  }
  // Filter on MAC ADDRESS
  if (setsockopt(recv_sockfd, SOL_SOCKET, SO_ATTACH_FILTER, &bpf, sizeof(bpf))) {
    perror("setsockopt SO_ATTACH_FILTER: ");
    return 1;
  }


  connect_sk_addr.sll_ifindex = if_index;
  connect_sk_addr.sll_halen = ETH_ALEN;
  for(int i = 0; i < 6; i++)
    connect_sk_addr.sll_addr[i] = dst_mac[i];

  memset(&standard_iq_packet, 0, sizeof(ecpri_iq_packet));
  for(int i = 0; i < 6; i++)
    standard_iq_packet.mac_src[i] = src_mac[i];
  for(int i = 0; i < 6; i++)
    standard_iq_packet.mac_dst[i] = dst_mac[i];

  /* Ethertype field */
  standard_iq_packet.ether_type = htons(0xaefe);

  /* Common Header */
  standard_iq_packet.ecpri_version = 0x10; // Version 0x1, Reserved = 0, C = 0
  standard_iq_packet.ecpri_type    = 0x00; // Message type (IQ data)
  standard_iq_packet.seq_id_fixed  = 0x80;

  /* ORAN HEADER */
  standard_iq_packet.oran_header[0] = 0x90;
  standard_iq_packet.oran_header[3] = 0x00;
  standard_iq_packet.oran_header[4] = 0x00;
  standard_iq_packet.oran_header[5] = 0x00;
  standard_iq_packet.oran_header[6] = 0x00;

  if(s->iq_file_trx) {
      read_ptr = fopen("/root/ecpri-iq-file.bin","rb");

      fseek(read_ptr, 0L, SEEK_END);
      iq_file_size = ftell(read_ptr);
      rewind(read_ptr);

      if(iq_file_size % (32 * sizeof(float))) {
        log_info("INIT_WRITE_DATA", "IQ file size not a multiple of 32 (%d)", iq_file_size);
        exit(EXIT_FAILURE);
      }
      log_info("INIT_WRITE_DATA", "IQ file size: %d", iq_file_size);

      fread(iq_file_bytes, iq_file_size, 1, read_ptr);
  }

  start_threads(s);
  return 0;
}

static void trx_ecpri_end(TRXState *s1)
{
  log_info("TRX_ECPRI", "End");
  TRXEcpriState *s = s1->opaque;
  free(s);
}

static int64_t prev_timestamp = 0;
static int64_t prev_count = 0;

/*
   Callback for processing TRX samples from Amarisoft
   Write count samples from __samples into TRX write buffer, exit as soon as possible

   - count: how much IQ samples to write (multiple of 32)
   - timestamp: should be equal to the amount of written IQ samples
   - __samples: can be null if zeroes are to be written (TDD only)
 */
static void trx_ecpri_write(TRXState *s1, trx_timestamp_t timestamp, const void **__samples, int count, int tx_port_index, TRXWriteMetadata *md)
{
  int64_t count_left, nc, offset;
  float ** _samples = (float **) __samples;
  int64_t a;
  int nc_sample, nc1, nc2;
  TRXEcpriState *s = s1->opaque;

#ifdef DISABLE_WRITE
  while(1)
    usleep(1000);
#endif
  /* Check for empty samples
  int k = 0;
  for(int j = 0; j < count; j++) {
    if(_samples[0][j] == 0)
      k++;
    else {
      k = 0;
      break;
    }
    if(k > 20)
      break;
  }
  if(k > 20) {
    printf("samples = \n");
    for(int j = 0; j < count / 2; j++)
      printf("%f ", _samples[0][j]);
    printf("\n");
    log_exit("TRX_ECPRI_WRITE", "Null data");
  }
  */

  //log_debug("TRX_ECPRI_WRITE", "trx_ecpri_write, count = %ld", count / N_SAMPLES);

  // Consistency check
  if(prev_count && ((timestamp - prev_timestamp) != prev_count)) {
    log_exit("TRX_ECPRI_WRITE",
        "Gap between timestamps: prev_ts %li ts %li prev_count %li count %li diff_ts %li",
        prev_timestamp, timestamp, prev_count, count, (timestamp - prev_timestamp));
  }
  prev_timestamp = timestamp; prev_count = count;

  // Drop samples if we don't have enough space in trx write buffer
  if( (count * 2) > rbuf_write_amount(&trxw_rbuf[0]) ) {
    update_counter(&tx_drop_counter, count);
    return;
  }

  for(int i = 0; i < s->tx_n_channel; i++) {
    offset = 0;
    count_left = count;
    while((nc = rbuf_contiguous_copy(NULL, &trxw_rbuf[i], count_left * 2))) {
      if(__samples) {
        if(s->iq_file_trx &&
           (write_counter.counter + (count - count_left) - nc / 2) >=  (1104 * 7 * 2 * 10 * 256 * 2)) {
          nc_sample = ((1104 * 7 * 2 * 10 * 256) - (write_counter.counter + (count - count_left))) * 2;
          nc1 = nc - nc_sample;
          nc2 = 0;
          memcpy((uint8_t *) trx_write_float_tab, ((uint8_t *) _samples[i]) + offset, nc_sample * 2);
          if((iq_file_size - iq_file_index) < (nc1 * 2)) {
            nc2 = nc1 - (iq_file_size - iq_file_index);
            nc1 = iq_file_size - iq_file_index;
          }
          memcpy((uint8_t *) trx_write_float_tab, iq_file_bytes + iq_file_index, nc1 * 2);
          iq_file_index = (iq_file_index + nc1 * 2) % iq_file_size;
          memcpy((uint8_t *) trx_write_float_tab, iq_file_bytes + iq_file_index, nc2 * 2);
          iq_file_index = (iq_file_index + nc2 * 2) % iq_file_size;
        } else {
          memcpy((uint8_t *) trx_write_float_tab, ((uint8_t *) _samples[i]) + offset, nc * 2);
        }
        float_to_int16(trx_write_int_tab, trx_write_float_tab, nc / 2, 32767);
        for(int k = 0; k < nc / 2; k++)
          trx_write_int_tab[k] = htons(trx_write_int_tab[k]);
        memcpy(rbuf_write(&trxw_rbuf[i]), (uint8_t *) trx_write_int_tab, nc);
      }
      else
        log_exit("TRX_ECPRI_WRITE", "samples empty during FDD mode");
      rbuf_increment_write(&trxw_rbuf[i], nc);
      count_left -= nc / 2;
      offset += nc * 2;
    }
  }
  update_counter(&write_counter, (s->tx_n_channel * count));
}

float mult = 1. / 32767.;

static int read_trxr(ring_buffer_t * rbuf, float * samples, int count_left, int offset) {
  int to_read = 0;
  int read = 0;
  int nc;
  int ra = rbuf_read_amount(rbuf);
  
  if(ra < (count_left * 2))
    to_read = ra;
  else
    to_read = count_left * 2;
  
  while((nc = rbuf_contiguous_copy(rbuf, NULL, to_read))) {
    memcpy((uint8_t*) trx_read_int_tab, rbuf_read(rbuf), nc);

    int mismatch = 0;
    //for(int k = 0; k < nc / 2; k++) {
    //  if(((uint8_t*) trx_read_int_tab)[2 * k] != 0x62 || ((uint8_t*)trx_read_int_tab)[2 * k +1] != 0x8f) {
    //    printf("2 * k = %d (%d) - %x%x \n", 2 * k, nc/2, ((uint8_t*)trx_read_int_tab)[2 * k], ((uint8_t*)trx_read_int_tab)[2 * k+1]);
    //    mismatch++;
    //    break;
    //  }
    //}
    //if(mismatch) {
    //  printf("%d/%d mismatch\n", mismatch, nc/2);
    //  for(int k = 0; k < nc / 2; k++) {
    //    printf("%x%x ", ((uint8_t*)trx_read_int_tab)[2 * k], ((uint8_t*)trx_read_int_tab)[2 * k+1]);
    //  }
    //  printf("\n");
    //  printf("ra = %d \n", rbuf_read_amount(rbuf));
    //  exit(EXIT_FAILURE);
    //}

    for(int k = 0; k < nc / 2 ; k++)
      trx_read_int_tab[k] = ntohs(trx_read_int_tab[k]);
    int16_to_float(trx_read_float_tab, trx_read_int_tab, nc / 2, mult);
    memcpy(((uint8_t*) samples) + offset + read, (uint8_t*) trx_read_float_tab, nc * 2);

    //mismatch = 0;
    //for(int k = 0; k < nc / 2; k++) {
    //  float diff = trx_read_float_tab[k] - 0.77f;
    //  diff = (diff > 0) ? diff : (diff * -1.f);
    //  if(diff > 0.01)
    //    mismatch++;
    //}
    //if(mismatch)
    //  printf("%d/%d mismatch\n", mismatch, nc/2);
    rbuf_increment_read(rbuf, nc);
    to_read -= nc;
    read += 2 * nc;
  }
  return read;
}

/*
   Callback for sending TRX samples to Amarisoft
   Put count IQ samples from TRX read buffer in __samples, return as soon as possible

   - count: how much IQ samples to read (multiple of 32)
   - ptimestamp: number of total read IQ samples
   - __samples: IQ samples returned
 */
static int trx_ecpri_read(TRXState *s1, trx_timestamp_t *ptimestamp, void **__samples,
    int count, int rx_port_index, TRXReadMetadata *md)
{
  int64_t count_left, offset, read;
  float ** _samples = (float **) __samples;
  TRXEcpriState *s = s1->opaque;

#ifdef DISABLE_READ
  while(1)
    usleep(1000);
#endif

  //log_debug("TRX_ECPRI_READ", "count = %d\n", count);

  for(int i = 0; i < s->rx_n_channel; i++ ) {
    offset = 0;
    count_left = count;
    while(count_left) {
      read = read_trxr(&trxr_rbuf[i], _samples[i], count_left, offset);
      //log_debug("TRX_ECPRI_READ", "read = %d\n", read);
#ifndef DISABLE_RECV2
      if(!read)
        read = read_trxr(&trxr2_rbuf[i], _samples[i], count_left, offset);
      if(!read)
        read = read_trxr(&trxr3_rbuf[i], _samples[i], count_left, offset);
      if(!read)
        read = read_trxr(&trxr4_rbuf[i], _samples[i], count_left, offset);
#endif
      if(!read) {
        usleep(10);
        continue;
      }
      count_left -= read / 4;
      offset += read;
    }
  }
  *ptimestamp = read_counter.counter / s->rx_n_channel;
  update_counter(&read_counter, s->rx_n_channel * count);

  //log_debug("TRX_ECPRI_READ", "read %d\n", count);

  return count;
}

/* This function can be used to automatically set the sample
   rate. Here we don't implement it, so the user has to force a given
   sample rate with the "sample_rate" configuration option */
static int trx_ecpri_get_sample_rate(TRXState *s1, TRXFraction *psample_rate,
    int *psample_rate_num, int sample_rate_min)
{
  log_info("TRX_ECPRI_GET_SAMPLE_RATE", "");
  return -1;
}

static int trx_ecpri_start(TRXState *s1, const TRXDriverParams *params)
{
  TRXEcpriState *s = s1->opaque;

  log_info("TRX_ECPRI_START", "Start");

  log_info("TRX_ECPRI_START", "trx_api_version: %d", s1->trx_api_version);
  log_info("TRX_ECPRI_START", "config file: %s", s1->path);
  log_info("TEST-ECPRI", "bbu-mac: %s, rrh-mac: %s, bbu-if: %s", s->bbu_mac, s->rrh_mac, s->bbu_if);

  s->sample_rate = params->sample_rate[0].num / params->sample_rate[0].den;

  start(s);

  return 0;
}

/* Called to start the tranceiver. Return 0 if OK, < 0 if  */
int trx_start_func(TRXState *s, const TRXDriverParams *p) {
  log_info("DEBUG", "trx_start_func");
  return 0;
}

/* Deprecated, use trx_write_func2 instead.
   Write 'count' samples on each channel of the TX port
   'tx_port_index'. samples[0] is the array for the first
   channel. timestamp is the time (in samples) at which the first
   sample must be sent. When the TRX_WRITE_FLAG_PADDING flag is
   set, samples is set to NULL. It indicates that no data should
   be sent (TDD receive time). TRX_WRITE_FLAG_END_OF_BURST is set
   to indicate in advance that the next write call will have the
   TRX_WRITE_FLAG_PADDING flag set. Note:
   TRX_WRITE_FLAG_END_OF_BURST and TRX_WRITE_FLAG_PADDING are
   never set simultaneously.
 */
void trx_write_func(TRXState *s, trx_timestamp_t timestamp, const void **samples, int count, int flags, int tx_port_index) {
  log_info("DEBUG", "**");
}

/* Deprecated, use trx_read_func2 instead.
   Read 'count' samples from each channel. samples[0] is the array
   for the first channel. *ptimestamp is the time at which the
   first samples was received. Return the number of sample read
   (=count).

Note: It is explicitely allowed that the application calls
trx_write_func, trx_read_func, trx_set_tx_gain_func and
trx_set_rx_gain_func from different threads.
 */
int trx_read_func(TRXState *s, trx_timestamp_t *ptimestamp, void **samples, int count, int rx_port_index) {
  log_info("DEBUG", "**");
  return 0;
}

/* Dynamic set the transmit gain (in dB). The origin and range are
   driver dependent.

Note: this function is only used for user supplied dynamic
adjustements.
 */
void trx_set_tx_gain_func(TRXState *s, double gain, int channel_num) {
  log_info("DEBUG", "trx_set_tx_gain_func");
}

/* Dynamic set the receive gain (in dB). The origin and range are
   driver dependent.

Note: this function is only used for user supplied dynamic
adjustements.
 */
void trx_set_rx_gain_func(TRXState *s, double gain, int channel_num) {
  log_info("DEBUG", "trx_set_rx_gain_func");
}

/* Return the maximum number of samples per TX packet. Called by
 * the application after trx_start_func.
 * Optional
 */
int trx_get_tx_samples_per_packet_func(TRXState *s) {
  log_info("DEBUG", "trx_get_tx_samples_per_packet_func");
  return 0;
}

/* Return some statistics. Return 0 if OK, < 0 if not available. */
int trx_get_stats(TRXState *s, TRXStatistics *m) {
  log_info("DEBUG", "trx_get_stats");
  return 0;
}

/* Callback must allocate info buffer that will be displayed */
void trx_dump_info(TRXState *s, trx_printf_cb cb, void *opaque) {
  log_info("DEBUG", "*opaque");
}

/* Return the absolute TX power in dBm for the TX channel
   'channel_num' assuming a square signal of maximum
   amplitude. This function can be called from any thread and
   needs to be fast. Return 0 if OK, -1 if the result is not
   available. */
int trx_get_abs_tx_power_func(TRXState *s,
    float *presult, int channel_num) {
  log_info("DEBUG", "trx_get_abs_tx_power_func");
  return 0;
}

/* Return the absolute RX power in dBm for the RX channel
   'channel_num' assuming a square signal of maximum
   amplitude. This function can be called from any thread and
   needs to be fast. Return 0 if OK, -1 if the result is not
   available. */
int trx_get_abs_rx_power_func(TRXState *s,
    float *presult, int channel_num) {
  log_info("DEBUG", "trx_get_abs_rx_power_func");
  return 0;
}

/* Remote API communication
 * Available since API v14
 * trx_msg_recv_func: called for each trx received messages
 * trx_msg_send_func: call it to send trx messages (They must be registered by client)
 * For each message, a call to send API must be done
 */
void trx_msg_recv_func(TRXState *s, TRXMsg *msg) {
  log_info("DEBUG", "trx_msg_recv_func");
}

TRXMsg* trx_msg_send_func(TRXState *s) {
  log_info("DEBUG", "trx_msg_send_func");
  return NULL;
}

/* Return actual transmit gain (in dB). The origin and range are
   driver dependent.
 */
void trx_get_tx_gain_func(TRXState *s, double *gain, int channel_num) {
  log_info("DEBUG", "trx_get_tx_gain_func");
}

/* Returns actual receive gain (in dB). The origin and range are
   driver dependent.
 */
void trx_get_rx_gain_func(TRXState *s, double *gain, int channel_num) {
  log_info("DEBUG", "trx_get_rx_gain_func");
}

/* Stop operation of the transceiver - to be called after trx_start.
   resources allocated in init are not released, so trx_call can be called again */
void trx_stop_func(TRXState *s) {
  log_info("DEBUG", "trx_stop_func");
}

/* OFDM mode: experimental 7.2 API */

/* read the current timestamp (only used in OFDM mode). Return 0
   if OK, < 0 if not supported by device. */
int trx_read_timestamp(TRXState *s, trx_timestamp_t *ptimestamp,
    int port_index) {
  log_info("DEBUG", "trx_read_timestamp");
  return 0;
}

void trx_set_tx_streams(TRXState *s, int rf_port_index,
    int n_streams, const TRXOFDMStreamInfo *streams) {
  log_info("DEBUG", "trx_set_tx_streams");
}
void trx_set_rx_streams(TRXState *s, int rf_port_index,
    int n_streams, const TRXOFDMStreamInfo *streams) {
  log_info("DEBUG", "trx_set_rx_streams");
}
/* schedule the reading of OFDM symbols. Return 0 if OK, < 0 if
   error. */
int trx_schedule_read(TRXState *s, int rf_port_index,
    const TRXScheduledSymbol *symbols,
    int n_symbols) {
  log_info("DEBUG", "trx_schedule_read");
  return 0;
}

/* AGC functions */
int trx_set_agc_func(TRXState *s, const TRXAGCParams *p, int channel) {
  log_info("DEBUG", "trx_set_agc_func");
  return 0;
}
int trx_get_agc_func(TRXState *s, TRXAGCParams *p, int channel) {
  log_info("DEBUG", "trx_get_agc_func");
  return 0;
}

int trx_driver_init(TRXState *s1)
{
  TRXEcpriState *s;
  double val;

  // Lock all current and future pages from preventing of being paged to swap
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    log_error("TRX_ECPRI", "mlockall failed");
  }

  log_info("TRX_ECPRI", "Init");

  if (s1->trx_api_version != TRX_API_VERSION) {
    fprintf(stderr, "ABI compatibility mismatch between LTEENB and TRX driver (LTEENB ABI version=%d, TRX driver ABI version=%d)\n",
        s1->trx_api_version, TRX_API_VERSION);
    return -1;
  }

  s = malloc(sizeof(TRXEcpriState));
  memset(s, 0, sizeof(*s));

  trx_get_param_double(s1, &val, "recv_affinity");
  s->recv_affinity              = (int) val;
  trx_get_param_double(s1, &val, "recv2_affinity");
  s->recv2_affinity              = (int) val;
  trx_get_param_double(s1, &val, "recv3_affinity");
  s->recv3_affinity              = (int) val;
  trx_get_param_double(s1, &val, "recv4_affinity");
  s->recv4_affinity              = (int) val;
  trx_get_param_double(s1, &val, "send_affinity");
  s->send_affinity              = (int) val;
  trx_get_param_double(s1, &val, "statistic_affinity");
  s->statistic_affinity         = (int) val;
  trx_get_param_double(s1, &val, "flow_id");
  s->flow_id                    = (int) val;
  trx_get_param_double(s1, &val, "trxw_buf_size");
  s->trxw_buf_size               = (int) val;
  trx_get_param_double(s1, &val, "trxr_buf_size");
  s->trxr_buf_size               = (int) val;
  trx_get_param_double(s1, &val, "txrx_buf_size");
  s->txrx_buf_size              = (int) val;
  trx_get_param_double(s1, &val, "rx_burst");
  s->rx_burst                   = (int) val;
  trx_get_param_double(s1, &val, "tx_burst");
  s->tx_burst                   = (int) val;
  trx_get_param_double(s1, &val, "rx_n_channel");
  s->rx_n_channel               = (int) val;
  trx_get_param_double(s1, &val, "tx_n_channel");
  s->tx_n_channel               = (int) val;
  trx_get_param_double(s1, &val, "statistics_refresh_rate_ms");
  s->statistics_refresh_rate_ms = (int) val;
  trx_get_param_double(s1, &val, "rx_drop_pcm");
  s->rx_drop_pcm         = (int) val;
  trx_get_param_double(s1, &val, "align_rx");
  s->align_rx            = (int) val;
  trx_get_param_double(s1, &val, "iq_file_trx");
  s->iq_file_trx         = (int) val;
  trx_get_param_double(s1, &val, "disable_rx");
  s->disable_rx            = (int) val;

  if(s->rx_n_channel == 0)
    log_exit("TRX_ECPRI", "rx_n_channel parameter can't be null\n");
  if(s->tx_n_channel == 0)
    log_exit("TRX_ECPRI", "tx_n_channel parameter can't be null\n");

  s->rrh_mac = trx_get_param_string(s1, "rrh_mac");
  s->bbu_mac = trx_get_param_string(s1, "bbu_mac");
  s->bbu_if = trx_get_param_string(s1, "bbu_if");
  s->log_directory = trx_get_param_string(s1, "log_directory");

  s1->opaque = s;
  s1->trx_end_func = trx_ecpri_end;
  s1->trx_write_func2 = trx_ecpri_write;
  s1->trx_read_func2 = trx_ecpri_read;
  s1->trx_start_func = trx_ecpri_start;
  s1->trx_get_sample_rate_func = trx_ecpri_get_sample_rate;

  //s1->trx_write_func = trx_write_func;
  //s1->trx_read_func = trx_read_func;
  //s1->trx_set_tx_gain_func = trx_set_tx_gain_func;
  //s1->trx_set_rx_gain_func = trx_set_rx_gain_func;
  //s1->trx_get_tx_samples_per_packet_func = trx_get_tx_samples_per_packet_func;
  //s1->trx_get_stats = trx_get_stats;
  //s1->trx_dump_info = trx_dump_info;
  //s1->trx_get_abs_tx_power_func = trx_get_abs_tx_power_func;
  //s1->trx_get_abs_rx_power_func = trx_get_abs_rx_power_func;
  //s1->trx_msg_recv_func = trx_msg_recv_func;
  //s1->trx_get_tx_gain_func = trx_get_tx_gain_func;
  //s1->trx_get_rx_gain_func = trx_get_rx_gain_func;
  //s1->trx_stop_func = trx_stop_func;
  //s1->trx_read_timestamp = trx_read_timestamp;
  //s1->trx_set_tx_streams = trx_set_tx_streams;
  //s1->trx_set_rx_streams = trx_set_rx_streams;
  //s1->trx_schedule_read = trx_schedule_read;
  //s1->trx_set_agc_func = trx_set_agc_func;
  //s1->trx_get_agc_func = trx_get_agc_func;
  //s1->trx_msg_send_func = trx_msg_send_func;

  return 0;
}
